/*
 *  SSL-Explorer
 *
 *  Copyright (C) 2003-2006 3SP LTD. All Rights Reserved
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
			
package com.sslexplorer.activedirectory;

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.regex.Pattern;

import org.apache.commons.cache.BaseStorageListener;
import org.apache.commons.cache.Cache;
import org.apache.commons.cache.FileStash;
import org.apache.commons.cache.GroupMapImpl;
import org.apache.commons.cache.LRUEvictionPolicy;
import org.apache.commons.cache.MemoryStash;
import org.apache.commons.cache.SimpleCache;
import org.apache.commons.cache.Stash;
import org.apache.commons.cache.StorageListener;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.sslexplorer.boot.ContextHolder;
import com.sslexplorer.boot.Util;
import com.sslexplorer.core.BundleActionMessage;
import com.sslexplorer.core.CoreUtil;
import com.sslexplorer.core.GlobalWarning;
import com.sslexplorer.policyframework.Principal;

public class PrincipalContainer<T extends Principal> {
    private static final Log logger = LogFactory.getLog(PrincipalContainer.class);
    private final int cacheSize;
    private final boolean inMemoryCache;
    private final String cacheType;
    private final Cache principalCache; 
    
    PrincipalContainer(int cacheSize, boolean inMemoryCache, String cacheType, String cacheFullMessageText) {
        this.cacheSize = cacheSize;
        this.inMemoryCache = inMemoryCache;
        this.cacheType = cacheType;
        principalCache = createCache(cacheFullMessageText);
    }
    
    synchronized boolean containsPrincipal(String principalName) {
        return retrievePrincipal(principalName) != null;
    }
    
    @SuppressWarnings("unchecked")
    synchronized T retrievePrincipal(String principalName) {
        return (T) principalCache.retrieve(principalName);
    }
    
    synchronized Collection<String> retrievePrincipalNames() {
        Serializable[] keysForGroup = principalCache.getKeysForGroup(cacheType);
        Collection<String> principalNames = new ArrayList<String>(keysForGroup.length);
        for (Serializable principalName : keysForGroup) {
            principalNames.add((String) principalName);
        }
        return principalNames;
    }
    
    @SuppressWarnings("unchecked")
    synchronized Collection<T> retrievePrincipals(String filter) {
        String regex = Util.parseSimplePatternToRegExp(filter);
        Pattern pattern = Pattern.compile(regex);

        Serializable[] keysForGroup = principalCache.getKeysForGroup(cacheType);
        Collection<T> principals = new HashSet<T>(keysForGroup.length);
        for (Serializable principalName : keysForGroup) {
            boolean matches = pattern.matcher((String) principalName).matches();
            if (matches) {
                T retrieve = (T) principalCache.retrieve(principalName);
                if (retrieve != null) {
                    principals.add(retrieve);
                }
            }
        }
        
        if (logger.isDebugEnabled()) {
            if (principals.isEmpty()) {
                logger.debug("No principals in cache");
            } else {
                logger.debug("Got " + principals.size() + " principals from the cache");                
            }
        }
        return principals;
    }

    synchronized String storePrincipal(T principal) {
        String principalName = principal.getPrincipalName();
        storePrinciple(principalName, principal);
        return principalName;
    }
    
    synchronized String storePrinciple(String key, T principal) {
        if (logger.isDebugEnabled()) {
            logger.debug("Caching " + principal);
        }
        principalCache.store(key, (Serializable) principal, Long.MAX_VALUE, null, cacheType);
        return key;
    }
    
    synchronized void updateRemovedPrincipals(Collection<String> missingPrincipals) {
        for (String principalName : missingPrincipals) {
            removePrincipal(principalName);
        }
    }
    
    public synchronized void removePrincipal(T principal) {
        removePrincipal(principal.getPrincipalName());
    }
    
    void removePrincipal(String principleName) {
        principalCache.store(principleName, null, 0L, null, cacheType);
    }
    
    synchronized void close() {
        closeCache(principalCache);
    }
    
    Cache createCache(String cacheFullMessageText) {
        File cacheDirectory = new File(ContextHolder.getContext().getTempDirectory(), "cache");
        File cacheTypeDirectory = new File(cacheDirectory, cacheType);
        Stash stash = inMemoryCache ? new MemoryStash(cacheSize) : new FileStash(Long.MAX_VALUE, cacheSize, new File[]{cacheTypeDirectory}, true);
        SimpleCache cache = new SimpleCache(stash, new LRUEvictionPolicy(), null, new GroupMapImpl());
        cache.registerStorageListener(getStorageListener(cacheFullMessageText));
        return cache;
    }

    private StorageListener getStorageListener(final String messageKey) {
        return new BaseStorageListener() {
            private static final long serialVersionUID = 4283488241230531541L;
            private int storageCounter = 0;
            private boolean addedWarning;

            public synchronized void stored(Serializable arg0, Serializable arg1, Long arg2, Long arg3, Serializable arg4) {
                storageCounter++;
            }

            public synchronized void cleared(Serializable arg0) {
                if (storageCounter == cacheSize && !addedWarning) {
                    BundleActionMessage message = new BundleActionMessage("activeDirectory", messageKey, String.valueOf(cacheSize));
                    CoreUtil.addMultipleGlobalWarning(GlobalWarning.MANAGEMENT_USERS, message);
                    addedWarning = true;
                }
                storageCounter--;
            }

            public synchronized void cleared() {
                storageCounter = 0;
                addedWarning = false;
                CoreUtil.removeGlobalWarningFromAllSessions(messageKey);
            }
        };
    }
    
    static void closeCache(Cache cache) {
        try {
            cache.clear();
            cache.unregisterStorageListeners();
        } catch (Exception e) {
            logger.error("Failed to close cache", e);
        }
    }
}