package com.sslexplorer.agent;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.maverick.multiplex.Channel;
import com.maverick.multiplex.ChannelFactory;
import com.maverick.multiplex.MultiplexedConnection;
import com.maverick.multiplex.Request;
import com.maverick.multiplex.RequestHandler;
import com.maverick.multiplex.TimeoutCallback;
import com.maverick.util.ByteArrayWriter;
import com.sslexplorer.boot.RequestHandlerTunnel;
import com.sslexplorer.boot.VersionInfo;
import com.sslexplorer.properties.Property;
import com.sslexplorer.properties.impl.profile.ProfilePropertyKey;
import com.sslexplorer.security.LogonControllerFactory;
import com.sslexplorer.security.SessionInfo;
import com.sslexplorer.security.User;

public class AgentTunnel extends MultiplexedConnection implements
		RequestHandlerTunnel, TimeoutCallback {

	static Log log = LogFactory.getLog(AgentTunnel.class);

	/**
	 * Request agent updates it resources
	 */
	public static final String UPDATE_RESOURCES_REQUEST = "updateResources@3sp.com";

	/**
	 * Synchronized
	 */
	public static final String SYNCHRONIZED_REQUEST = "synchronized@3sp.com";

	MultiplexedConnection con = null;

	String type;

	HashMap properties = new HashMap();

	String id;

	SessionInfo session;

	SyncHandler syncHandler = new SyncHandler();

	int timeoutId;
	public static final int KEEP_ALIVE_TIMEOUT = 10000;

	/**
	 * Constructor for agent tunnels that are <strong>not</strong> associated
	 * with a UI session.
	 * 
	 * @param id
	 *            agent ID
	 * @param user
	 * @param type
	 * @param factory
	 */
	public AgentTunnel(String id, User user, String type, ChannelFactory factory) {
		this(id, null, user, type, factory);
	}

	/**
	 * Constructor for agent tunnels that may or may not be associated with a UI
	 * session
	 * 
	 * @param id
	 *            agent ID
	 * @param session
	 *            UI session the tunnel is associated with or <code>null</code>
	 *            if if not associated with a user session
	 * @param user
	 * @param type
	 * @param factory
	 */
	public AgentTunnel(String id, SessionInfo session, User user, String type,
			ChannelFactory factory) {
		super(factory);
		this.id = id;
		this.session = session;
		this.type = type;
		registerRequestHandler(AgentTunnel.SYNCHRONIZED_REQUEST, syncHandler);
	}

	/**
	 * Get the agent ID
	 * 
	 * @return agent ID
	 */
	public String getId() {
		return id;
	}

	public void register() {
		timeoutId = -1;
		if (session.getHttpSession() != null) {
			if (Property.getPropertyBoolean(new ProfilePropertyKey(
					"client.preventSessionTimeoutIfActive", session))) {
				timeoutId = LogonControllerFactory.getInstance()
						.addSessionTimeoutBlock(session.getHttpSession(),
								"SSL-Explorer Agent");
			}
		}

	}

	public void unregister() {
		if (!session.isInvalidating() && timeoutId != -1) {
			LogonControllerFactory.getInstance().removeSessionTimeoutBlock(
					session.getHttpSession(), timeoutId);
		}
	}

	/**
	 * Get the UI session the agent tunnel is associated with or
	 * <code>null</code> if it is not associated with a tunnel.
	 * 
	 * @return session
	 */
	public SessionInfo getSession() {
		return session;
	}

	public String getType() {
		return type;
	}

	public Object getProperty(String name) {
		return properties.get(name);
	}

	public void setProperty(String name, Object value) {
		properties.put(name, value);
	}

	public void tunnel(InputStream _in, OutputStream _out) {

		if (log.isDebugEnabled())
			log.debug("Starting agent tunnel");

		/**
		 * LDP - Configure a timeout callback on the multiplexed connection
		 */
		setTimeoutCallback(this);

		startProtocol(_in, _out, false);

		if (log.isDebugEnabled())
			log.debug("Stopping agent tunnel");
        
        // Inform all channels they are closing
        Channel[] channels = getActiveChannels();
        for(int i = 0 ; i < channels.length; i++) {
            channels[i].onChannelClose();
        }

		try {
			AgentManager.getInstance().removeAgent(this);
		} catch (AgentException e) {
			log.error("Error removing agent", e);
		}

	}

	public boolean waitForSync(long ms) throws InterruptedException {
		synchronized (syncHandler) {
			if (syncHandler.isSyncCompleted()) {
				return true;
			}
			syncHandler.wait(ms);
			return syncHandler.isSyncCompleted();
		}
	}

	public class SyncHandler implements RequestHandler {

		private boolean syncCompleted;

		public boolean isSyncCompleted() {
			return syncCompleted;
		}

		public synchronized boolean processRequest(Request request,
				MultiplexedConnection connection) {
			try {
				notifyAll();
				ByteArrayWriter baw = new ByteArrayWriter();
				try {
					baw.writeString(VersionInfo.getVersion().toString());
					request.setRequestData(baw.toByteArray());
					return true;
				} catch (IOException e) {
					log.error("Failed to send back server version.");
					return false;
				}
			} finally {
				syncCompleted = true;
			}
		}

		public void postReply(MultiplexedConnection connection) {
			DefaultAgentManager.getInstance().startServices(
					(AgentTunnel) connection);
		}
	}
	
	public boolean isAlive(MultiplexedConnection con) {
		/**
		 * We will schedule a keep alive packet and disconnect if that
		 * fails. We do not perform that here as it would block the multiplex
		 * thread.
		 */
		Thread t = new Thread("Keep-Alive Thread - " + session.getUser().getPrincipalName() + "," + session.getAddress().getHostAddress()) {
			public void run() {
				Request request = new Request("keepAlive@3sp.com");
				if(log.isInfoEnabled())
					log.info("Socket has timed out! Sending keep-alive to agent");
				
				try {
					sendRequest(request, true, KEEP_ALIVE_TIMEOUT);
					
					if(log.isInfoEnabled())
						log.info("Agent is still alive, continue as normal");
					
				} catch (IOException e) {
					
					if(log.isInfoEnabled())
						log.info("Agent did not respond, disconnecting");
					
					disconnect("Keep-alive failed");
				}
			}
		};
		t.start();
		return true;
	}
}
